# -*- coding: utf-8 -*-
"""Test extraction functions in nfetl.processes.extract"""


# %% Imports
# %%% Py3 Standard
from typing import List

# %%% 3rd Party
import pytest
import pandas as pd

# %%% User Defined
from nfetl.processes.extract import get_url_data, get_update
from nfetl._config import _config as config
from nfetl._datetime import date


# %% Functions
def test_get_url_data():
    actual_data: pd.DataFrame = get_url_data(
        config['URLs']['test'], config['Columns_extract']['test'].split(', '))
    test_data: pd.DataFrame = pd.read_hdf(config['Data']['test_url_data'],
                                          'test_url_data')
    assert (actual_data.to_numpy() == test_data.to_numpy()).all()

def test_get_update():
    dfs: List[pd.DataFrame] = get_update(date.today().nfl_year)
    try:
        not_empty: List[bool] = [len(df.index) > 0 for df in dfs.values()]
    except AttributeError:
        not_empty: List[bool] = [False]
    urls: List[str] = [key for key in config['URLs'].keys()]
    assert all(not_empty) and len(dfs) == len(urls) - 1



# -*- coding: utf-8 -*-


# %% Imports
# %%% Py3 Standard
from configparser import ConfigParser

# %%% 3rd Party
import pytest

# %%% User Defined
from nfetl._config import _config as config


# %% Functions
def test_config() -> None:
    assert isinstance(config, ConfigParser) and config.sections() == [
        'Scope', 'Data', 'URLs', 'Columns_extract', 'Columns_transform']



# -*- coding: utf-8 -*-

# %% Imports
# %%% 3rd Party
import pytest

# %%% User Defined
from datetime import date
from nfetl._datetime import date as test_date


# %% Functions
@pytest.mark.parametrize("dtInput, intOutput",
                         [(test_date(date.today().year, 9, 8),
                           date.today().year),
                          (test_date(date.today().year, 2, 1),
                           date.today().year - 1)])
def test_nfl_year(dtInput, intOutput):
    assert dtInput.nfl_year == intOutput



# -*- coding: utf-8 -*-
"""Adds nfl_year attribute to standard python date object."""


# %% Imports
from datetime import date


# %% Classes
class _date(date):
    __doc__ = date.__doc__

    # %%% Functions
    @property
    def nfl_year(self):
        """
        Get nfl_year property.

        Returns
        -------
        int
            NFL season year.
        """
        if self.month in [10, 11, 12] or (self.month == 9 and
                                          self.day / 7 > 1):
            return self.year
        else:
            return self.year - 1

date = _date



# -*- coding: utf-8 -*-
"""Load configured variables."""


# %% Imports
from configparser import ConfigParser
import os


# %% Variables
_self_path: str = os.path.dirname(__file__)
_config_path: str = os.path.join(_self_path, 'config.txt')

_config = ConfigParser()
_config.read(_config_path)
_config['Data']['folder'] = os.path.join(os.path.dirname(_self_path), 'data')
with open(_config_path, 'w') as file:
    _config.write(file)



# -*- coding: utf-8 -*-
"""Transform data sourced from pro-football-reference.com."""


# %% Imports
import pandas as pd
# TODO: consider this
# import dask.dataframe as dd


# %% Functions


# %% Script
if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
"""Extract data from pro-football-reference.com."""


# %% Imports
# %%% Py3 Standard
from random import randrange, random
from time import sleep
import urllib.request
from urllib.error import URLError
from typing import List, Tuple, Dict

# %%% 3rd Party
from bs4 import BeautifulSoup
import pandas as pd
from dask import distributed

# %%% User Defined
from nfetl._config import _config as config
from nfetl._datetime import date


# %% Variables
# %%% System
__all__ = ['get_url_data', 'get_update']

# %%% Private
_client: object = distributed.Client(processes=False)

_default_start_year: int = int(config['Scope']['start_year'])
_default_sets: List[Tuple[str, str, str]] = [
    (key, config['URLs'][key], config['Columns_extract'][key])
    for key in config['Scope']['keys'].split(', ')]


# %% Functions
def get_url_data(url: str, headers: List[str] = []) -> pd.DataFrame:
    """
    Retrieve data table from URL.

    Parameters
    ----------
    url : str
        URL for HTML page where data table is found.
    headers : List[str], optional
        Headers for retrieved table. Default is source headers.

    Returns
    -------
    data : pandas.DataFrame
        Data table from HTML page found at URL.
    """
    try:
        html: object = urllib.request.urlopen(url)
    except URLError:
        return pd.DataFrame()
    table: List[str] = BeautifulSoup(html, 'html.parser').findAll('tr')
    # Reduce table size if more than one table is present in data
    i: int = 0
    for row in table:
        try:
            if i > 1 and 'thead' not in row.attrs.get('class')[0]:
                table = table[:i]
                break
        except TypeError:
            pass
        finally:
            i += 1
    # sub_table tracks the subsection being retrieved
    sub_table: List[str] = table
    # remaining_table tracks the table sections yet to be retrieved
    remaining_table: List[str] = table
    for header_row in [1, 0]:
        if len(headers) < 2:
            headers = [i.getText() for i in table[header_row].findAll('th')]
    data: pd.DataFrame = pd.DataFrame(columns=headers)
    # variables for building sub_table, remaining table
    done: bool = False
    started: bool = False
    i = 0
    b: int = i
    while not done:
        try:
            data = data.append(
                pd.DataFrame([max([c.getText() for c in row],
                                  ['']*len(headers)) for row in sub_table],
                             columns=headers))
            remaining_table = remaining_table[i:]
            sub_table = remaining_table
            if len(remaining_table) == 0:
                done = True
        except (ValueError, AttributeError) as e:
            if isinstance(e, AttributeError):
                i = 0
                b = 0
                for row in sub_table:
                    i += 1
                    if any([not hasattr(c, 'getText') for c in row]):
                        if started and i > b + 1:
                            i -= 1
                            break
                        else:
                            b = i
                    else:
                        if not started:
                            started = True
                sub_table = sub_table[b:i]
            else:
                i = 0
                for row in sub_table:
                    if len(max([c.getText() for c in row],
                               ['']*len(headers))) == len(headers):
                        i += 1
                    else:
                        continue
                sub_table = sub_table[:i]
    return data


def get_update(start_year: int = _default_start_year,
               datasets: List[Tuple[str, str, str]] = _default_sets):
    """
    Retrieve all data tables from URL list provided by user or in config file.

    Makes multiple calls to get_url_data from the dask distributed client. Will
    retrieve data from start_year and after.

    Parameters
    ----------
    start_year : int, optional
        Four digit year for start of update. Default is from config file.
    datasets : List[Tuple[str, str]], optional
        Contains URL strings and header lists from user input or config file.
        Default are from config file.

    Returns
    -------
    dfs : list
        Contains pandas.DataFrame objects containing data from table found on
        HTML page located at URL.
    """
    if start_year > date.today().nfl_year:
        raise Exception
    dfs: Dict[str, pd.DataFrame] = {}
    headers: List[str] = []
    url: str = ''
    for dataset in datasets:
        headers = dataset[2].split(', ')
        for y in range(start_year, date.today().nfl_year + 1):
            url = dataset[1].format(str(y))
            dfs[dataset[0]] = _client.submit(get_url_data, url=url, headers=headers)
            sleep(randrange(3, 6) + random())
    distributed.wait(dfs.values())
    dfs = {v[0]: v[1].result() for v in dfs.items()}
    return dfs



# -*- coding: utf-8 -*-
"""ETL Functions for NFL statistics database."""


# %% Imports
from nfetl.processes.extract import (get_url_data,
                                     get_update)
from nfetl.processes.transform import *
from nfetl.processes.load import *

# %% Variables
__all__ = ['get_url_data', 'get_update']



# -*- coding: utf-8 -*-
"""Setup nfetl package."""


from setuptools import setup, find_packages


with open("README.md", 'r') as f:
    long_description: str = f.read()

setup(
      name='nfetl',
      version="0.1.1",
      description="A package for creating a database containing NFL stats.",
      long_description=long_description,
      author="Dan Eschman",
      author_email="deschman007@gmail.com",
      license="GNU AGPLv3",
      python_requires='>=3.7',
      install_requires=['pytest', 'bs4', 'pandas', 'dask'],
      packages=find_packages(),
      package_data={"test_URLData": ['data/test_URLData.h5']},
      long_description_content_type='text/markdown',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU Affero General Public License v3',
          'Natural Language :: English',
          'Operating System :: Microsoft :: Windows',
          'Programming Language :: Python :: 3',
          'Topic :: Database',
          'Topic :: Games/Entertainment',
          'Topic :: Internet',
          'Topic :: Text Processing'])



# -*- coding: utf-8 -*-
"""Module defining internal database object, DB."""


# %% Imports
# %%% Py3 Standard
import sys
import os
import sqlite3
from datetime import date

# %%% User Defined
import nfetl


# %% Variables
# %%% System
__all__ = ['DB']

# %%% Private
_default_db_path: str = os.path.join(os.path.dirname(sys.argv[0]), 'NFL.db')


# %% Classes
class DB(object):
    r"""
    Base database object class.

    Parameters
    ----------
    db_path : str, optional
        Path where database will be located. The default is nfetl\db\NFL.db.
    """

    # %% Functions
    # %%% Private
    def __init__(self, db_path: str = _default_db_path) -> None:
        self._db_path = db_path
        self.__connection = sqlite3.connect(self.db_path)

    @property
    def _connection(self) -> sqlite3.Connection:
        return self.__connection

    @_connection.setter
    def _set_connection(self, new_connection: sqlite3.Connection) -> None:
        if self.__connection is not None:
            self.__connection.close()
        self.__connection = new_connection

    @_connection.deleter
    def _del_connection(self) -> None:
        self.__connection.close()

    # %%% Public
    @property
    def connection(self) -> sqlite3.Connection:
        """
        Get connection property.

        Returns
        -------
        sqlite3.Connection object
            Connection to database. If db_path property is changed, this
            connection will be reset.
        """
        return self._connection

    @property
    def db_path(self) -> str:
        """
        Get db_path property.

        Returns
        -------
        str
            File path to database.
        """
        return self._db_path

    @db_path.setter
    def _set_db_path(self, new_path: str) -> None:
        self._db_path = new_path
        self._connection = sqlite3.connect(self._db_path)

    @property
    def last_update(self) -> date:
        """
        Get last_update property.

        date object
            Date that database was updated.
        """
        return self._last_update

    def update(self) -> None:
        """Extract and load all data not currently in database."""
        data = nfetl.extract.get_update()
        data = nfetl.transform.clean(data)
        nfetl.load.add_data(data)
        nfetl.transform.refresh_db()


# %% Script
if __name__ == '__main__':
    if len(sys.argv) > 1:
        db = DB(sys.argv[1])
    else:
        db = DB()
    db.update()



# -*- coding: utf-8 -*-


# %% Imports
# %%% 3rd Pary
import pytest

# %%% User Defined
from nfetl.processes.extract import get_url_data, get_update
from nfetl._config import _config as config
from nfetl.processes.transform import *


# %% Functions


# %% Script
if __name__ == '__main__':
    pass
